from dataclasses import dataclass, field
from typing import Any, List, OrderedDict

from ..config import MappingConfigMapping, StairlightConfigInclude
from ..template import TemplateSourceType as source_type


@dataclass
class StairlightConfigIncludeFile(StairlightConfigInclude):
    TemplateSourceType: str = source_type.FILE.value
    FileSystemPath: str = None
    Regex: str = None
    DefaultTablePrefix: str = None


@dataclass
class MappingConfigMappingFile(MappingConfigMapping):
    TemplateSourceType: str = source_type.FILE.value
    FileSuffix: str = None
    Tables: List[OrderedDict[str, Any]] = field(default_factory=list)



from collections import OrderedDict as ordered_dict
from dataclasses import dataclass, field
from typing import Any, List, OrderedDict

from ..config import MappingConfigMapping, StairlightConfigInclude
from ..template import TemplateSourceType as source_type


@dataclass
class StairlightConfigIncludeDbt(StairlightConfigInclude):
    TemplateSourceType: str = source_type.DBT.value
    ProjectDir: str = None
    ProfilesDir: str = None
    Target: str = None
    Vars: OrderedDict[str, str] = field(default_factory=ordered_dict)


@dataclass
class MappingConfigMappingDbt(MappingConfigMapping):
    TemplateSourceType: str = source_type.DBT.value
    ProjectName: str = None
    FileSuffix: str = None
    Tables: List[OrderedDict[str, Any]] = field(default_factory=list)



import glob
import os
import pathlib
import re
import shlex
import subprocess
from typing import Any, Dict, Iterator, List

import yaml

from ..config import MappingConfig, StairlightConfig
from ..config_key import DbtProjectKey
from ..template import Template, TemplateSource, TemplateSourceType
from .config import StairlightConfigIncludeDbt


class DbtTemplate(Template):
    def __init__(
        self,
        mapping_config: MappingConfig,
        key: str,
        project_name: str,
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=key,
            source_type=TemplateSourceType.DBT,
        )
        self.uri = self.get_uri()
        self.project_name = project_name

    def get_uri(self) -> str:
        """Get uri from file path

        Returns:
            str: uri
        """
        return str(pathlib.Path(self.key).resolve())

    def get_template_str(self) -> str:
        """Get template string that read from a file

        Returns:
            str: Template string
        """
        with open(self.key) as f:
            return f.read()

    def render(
        self, params: Dict[str, Any] = None, ignore_params: List[str] = None
    ) -> str:
        return self.get_template_str()


class DbtTemplateSource(TemplateSource):
    DBT_PROJECT_YAML = "dbt_project.yml"
    REGEX_SCHEMA_TEST_FILE = re.compile(r".*/schema.yml/.*\.sql$")

    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        include: StairlightConfigIncludeDbt,
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config,
            mapping_config=mapping_config,
        )
        self._include = include

    def search_templates(self) -> Iterator[Template]:
        project_dir: str = self._include.ProjectDir
        dbt_project_config: Dict[str, Any] = self.read_dbt_project_yml(
            project_dir=project_dir
        )

        _ = self.execute_dbt_compile(
            project_dir=project_dir,
            profiles_dir=self._include.ProfilesDir,
            profile=dbt_project_config.get(DbtProjectKey.PROFILE),
            target=self._include.Target,
            vars=self._include.Vars,
        )

        for model_path in dbt_project_config[DbtProjectKey.MODEL_PATHS]:
            dbt_model_path_str = self.concat_dbt_model_path_str(
                project_dir=project_dir,
                dbt_project_config=dbt_project_config,
                model_path=model_path,
            )
            dbt_model_path = pathlib.Path(dbt_model_path_str)
            for obj in dbt_model_path.glob("**/*"):
                if (
                    (obj.is_dir())
                    or (self.REGEX_SCHEMA_TEST_FILE.fullmatch(str(obj)))
                    or self.is_excluded(
                        source_type=TemplateSourceType(
                            self._include.TemplateSourceType
                        ),
                        key=str(obj),
                    )
                ):
                    self.logger.debug(f"{str(obj)} is skipped.")
                    continue

                yield DbtTemplate(
                    mapping_config=self._mapping_config,
                    key=str(obj),
                    project_name=dbt_project_config[DbtProjectKey.PROJECT_NAME],
                )

    def read_dbt_project_yml(self, project_dir: str) -> dict:
        """Read dbt_project.yml

        Args:
            project_dir (str): dbt project directory

        Returns:
            dict: dbt project settings
        """
        dbt_project_pattern = re.compile(f"^{project_dir}/{self.DBT_PROJECT_YAML}$")
        return self.read_yml(dir=project_dir, re_pattern=dbt_project_pattern)

    def read_yml(self, dir: str, re_pattern: re.Pattern) -> dict:
        files = [
            obj
            for obj in glob.glob(f"{dir}/**", recursive=False)
            if re_pattern.fullmatch(obj)
        ]
        if not files:
            self.logger.error(f"Not found: {re_pattern} in {os.getcwd()}")
            exit()
        with open(files[0]) as file:
            return yaml.safe_load(file)

    @staticmethod
    def concat_dbt_model_path_str(
        project_dir: str,
        dbt_project_config: Dict[str, Any],
        model_path: pathlib.Path,
    ) -> str:
        return (
            f"{project_dir}/"
            f"{dbt_project_config[DbtProjectKey.TARGET_PATH]}/"
            "compiled/"
            f"{dbt_project_config[DbtProjectKey.PROJECT_NAME]}/"
            f"{model_path}/"
        )

    @staticmethod
    def execute_dbt_compile(
        project_dir: str,
        profiles_dir: str,
        profile: str = None,
        target: str = None,
        vars: Dict[str, Any] = None,
    ) -> int:
        command = (
            "dbt compile "
            f"--project-dir {project_dir} "
            f"--profiles-dir {profiles_dir} "
        )
        if profile:
            command += f"--profile {profile} "
        if target:
            command += f"--target {target} "
        if vars:
            command += f"--vars '{vars}' "

        proc = subprocess.run(
            args=shlex.split(command),
            shell=False,
            check=True,
            stdout=subprocess.DEVNULL,
        )
        return proc.returncode



import glob
import logging
import re
from collections import OrderedDict
from dataclasses import asdict
from datetime import datetime, timezone
from typing import Any, Dict, List

import yaml

from .source.config import (
    MappingConfig,
    MappingConfigGlobal,
    MappingConfigMapping,
    MappingConfigMappingTable,
    MappingConfigMetadata,
    StairlightConfig,
    StairlightConfigExclude,
    StairlightConfigSettings,
)
from .source.config_key import MapKey
from .source.controller import collect_mapping_attributes, get_default_table_name
from .source.dbt.config import StairlightConfigIncludeDbt
from .source.file.config import StairlightConfigIncludeFile
from .source.gcs.config import StairlightConfigIncludeGcs
from .source.redash.config import StairlightConfigIncludeRedash
from .source.s3.config import StairlightConfigIncludeS3
from .source.template import Template

logger = logging.getLogger()

STAIRLIGHT_CONFIG_PREFIX_DEFAULT = "stairlight"
MAPPING_CONFIG_PREFIX_DEFAULT = "mapping"


class Configurator:
    def __init__(self, dir: str) -> None:
        """Configuration class

        Args:
            path (str): Configuration file path
        """
        self.dir = dir

    def read_stairlight(
        self, prefix: str = STAIRLIGHT_CONFIG_PREFIX_DEFAULT
    ) -> StairlightConfig:
        config = self.read(prefix=prefix)
        return StairlightConfig(**config)

    def read_mapping(
        self, prefix: str = MAPPING_CONFIG_PREFIX_DEFAULT
    ) -> MappingConfig:
        config = self.read(prefix=prefix)
        return MappingConfig(**config)

    def read(self, prefix: str) -> Dict[str, Any]:
        """Read a configuration file

        Args:
            prefix (str): Configuration file name prefix

        Returns:
            dict: Results from reading configuration file
        """
        config: Dict[str, Any] = {}
        pattern = f"^{self.dir}/{prefix}.ya?ml$"
        config_file = [
            p
            for p in glob.glob(f"{self.dir}/**", recursive=False)
            if re.fullmatch(pattern, p)
        ]
        if config_file:
            with open(config_file[0]) as file:
                config = yaml.safe_load(file)
        return config

    def create_stairlight_file(
        self, prefix: str = STAIRLIGHT_CONFIG_PREFIX_DEFAULT
    ) -> str:
        """Create a Stairlight template file

        Args:
            prefix (str, optional): File prefix. Defaults to STAIRLIGHT_CONFIG_PREFIX.

        Returns:
            str: Created file name
        """
        template_file_name = f"{self.dir}/{prefix}.yaml"
        with open(template_file_name, "w") as f:
            yaml.add_representer(OrderedDict, self.represent_odict)
            yaml.dump(self.build_stairlight_config(), f)
        return template_file_name

    def create_mapping_file(
        self,
        unmapped: List[Dict[str, Any]],
        prefix: str = MAPPING_CONFIG_PREFIX_DEFAULT,
    ) -> str:
        """Create a mapping template file

        Args:
            unmapped (list[dict]): Unmapped results
            prefix (str, optional): File prefix. Defaults to MAPPING_CONFIG_PREFIX.

        Returns:
            str: Mapping template file
        """
        now = datetime.now(timezone.utc).strftime("%Y%m%d%H%M%S")
        template_file_name = f"{self.dir}/{prefix}_{now}.yaml"

        with open(template_file_name, "w") as f:
            yaml.add_representer(
                data_type=OrderedDict, representer=self.represent_odict
            )
            yaml.dump(
                self.build_mapping_config(unmapped_templates=unmapped),
                stream=f,
            )
        return template_file_name

    @staticmethod
    def represent_odict(
        dumper: yaml.Dumper, odict: OrderedDict
    ) -> yaml.nodes.MappingNode:
        """Create a OrderedDict object for dumping a YAML file
        in order of OrderedDict"""
        return dumper.represent_mapping(
            tag="tag:yaml.org,2002:map", mapping=odict.items()
        )

    @staticmethod
    def build_stairlight_config() -> OrderedDict:
        """Create a OrderedDict object for file 'stairlight.config'

        Returns:
            OrderedDict: stairlight.config template
        """
        return OrderedDict(
            asdict(
                StairlightConfig(
                    Include=[
                        OrderedDict(asdict(StairlightConfigIncludeFile())),
                        OrderedDict(asdict(StairlightConfigIncludeGcs())),
                        OrderedDict(asdict(StairlightConfigIncludeRedash())),
                        OrderedDict(asdict(StairlightConfigIncludeDbt())),
                        OrderedDict(asdict(StairlightConfigIncludeS3())),
                    ],
                    Exclude=[OrderedDict(asdict(StairlightConfigExclude()))],
                    Settings=OrderedDict(asdict(StairlightConfigSettings())),
                )
            ),
        )

    def build_mapping_config(
        self, unmapped_templates: List[Dict[str, Any]]
    ) -> OrderedDict:
        """Create a OrderedDict for mapping.yaml

        Args:
            unmapped_templates (list[dict]): unmapped settings that Stairlight detects

        Returns:
            OrderedDict: mapping.yaml template
        """
        # List(instead of Set) because OrderedDict is not hashable
        parameters_set: List[OrderedDict] = []
        global_parameters: Dict[str, Any] = {}

        # Mapping section
        mappings: List[MappingConfigMapping] = []
        unmapped_template: Dict[str, Any]
        for unmapped_template in unmapped_templates:
            # Tables.Parameters
            parameters: OrderedDict = OrderedDict()
            template: Template = unmapped_template[MapKey.TEMPLATE]

            if MapKey.PARAMETERS in unmapped_template:
                undefined_params: List[str] = unmapped_template.get(
                    MapKey.PARAMETERS, []
                )
                for undefined_param in undefined_params:
                    splitted_params = undefined_param.split(".")
                    create_nested_dict(keys=splitted_params, results=parameters)

                if parameters in parameters_set:
                    global_parameters.update(parameters)
                else:
                    parameters_set.append(parameters)

            table = OrderedDict(
                asdict(
                    MappingConfigMappingTable(
                        TableName=get_default_table_name(template=template),
                        Parameters=parameters,
                        IgnoreParameters=[],
                        Labels=OrderedDict(),
                    )
                )
            )
            mapping: MappingConfigMapping = collect_mapping_attributes(
                template=template,
                tables=[table],
            )
            mappings.append(mapping)

        return OrderedDict(
            asdict(
                MappingConfig(
                    Global=OrderedDict(
                        asdict(MappingConfigGlobal(Parameters=global_parameters))
                    ),
                    Mapping=[OrderedDict(asdict(mapping)) for mapping in mappings],
                    Metadata=[OrderedDict(asdict(MappingConfigMetadata()))],
                )
            )
        )


def create_nested_dict(
    keys: List[str],
    results: Dict[str, Any],
    density: int = 0,
    default_value: Any = None,
) -> None:
    """create nested dict from list

    Args:
        keys (list): Dict keys
        results (dict[str, Any]): Nested dict
        density (int, optional): Density. Defaults to 0.
        default_value (any, optional): Default dict value. Defaults to None.
    """
    key = keys[density]
    if density < len(keys) - 1:
        if key not in results:
            results[key] = {}
        create_nested_dict(keys=keys, results=results[key], density=density + 1)
    else:
        results[key] = default_value



from dataclasses import dataclass, field
from typing import Any, List, OrderedDict

from ..config import MappingConfigMapping, StairlightConfigInclude
from ..template import TemplateSourceType as source_type


@dataclass
class StairlightConfigIncludeS3(StairlightConfigInclude):
    TemplateSourceType: str = source_type.S3.value
    ProjectId: str = None
    BucketName: str = None
    Regex: str = None
    DefaultTablePrefix: str = None


@dataclass
class MappingConfigMappingS3(MappingConfigMapping):
    TemplateSourceType: str = source_type.S3.value
    Uri: str = None
    Tables: List[OrderedDict[str, Any]] = field(default_factory=list)



from logging import getLogger
from typing import Any, Dict, Iterator, List, OrderedDict, Type

from .query import Query
from .source.config import (
    MappingConfig,
    MappingConfigGlobal,
    MappingConfigMappingTable,
    StairlightConfig,
)
from .source.config_key import MapKey, MappingConfigKey
from .source.controller import get_template_source_class
from .source.template import Template, TemplateSource, TemplateSourceType

logger = getLogger(__name__)


class Map:
    """Manages functions related to dependency map objects"""

    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        mapped: Dict[str, Any] = None,
    ) -> None:
        """Manages functions related to dependency map objects

        Args:
            stairlight_config (StairlightConfig): Stairlight configuration
            mapping_config (MappingConfig): Mapping configuration
            mapped (dict, optional):
                Mapped file attributes when a mapping configuration file loaded.
                Defaults to {}.
        """
        if mapped:
            self.mapped = mapped
        else:
            self.mapped = {}
        self.unmapped: List[dict] = []
        self._stairlight_config = stairlight_config
        self._mapping_config = mapping_config

    def write(self) -> None:
        """Write a dependency map"""
        template_source: TemplateSource
        for template_source in self.find_template_source():
            self.write_by_template_source(template_source=template_source)

    def find_template_source(self) -> Iterator[TemplateSource]:
        """find template source

        Yields:
            Iterator[TemplateSource]: Template source instance
        """
        for include in self._stairlight_config.get_include():
            template_source: Type[TemplateSource] = get_template_source_class(
                template_source_type=include.TemplateSourceType
            )
            if not template_source:
                logger.warning(msg=f"Template source is not found: {type}")
                continue
            yield template_source(
                stairlight_config=self._stairlight_config,
                mapping_config=self._mapping_config,
                **{"include": include},
            )

    def write_by_template_source(self, template_source: TemplateSource) -> None:
        """Write a dependency map by template source"""
        for template in template_source.search_templates():
            if not self._mapping_config:
                self.add_unmapped_params(template=template)
            elif template.is_mapped():
                for table_attributes in template.find_mapped_table_attributes():
                    unmapped_params = self.detect_unmapped_params(
                        template=template, table_attributes=table_attributes
                    )
                    if unmapped_params:
                        self.add_unmapped_params(
                            template=template, params=unmapped_params
                        )
                    self.remap(template=template, table_attributes=table_attributes)
            else:
                self.add_unmapped_params(template=template)

    def remap(
        self, template: Template, table_attributes: MappingConfigMappingTable
    ) -> None:
        """Remap a dependency map

        Args:
            template (Template): SQL template
            table_attributes (MappingConfigMappingTable):
                Table attributes from mapping configuration
        """
        query_str: str = template.render(
            params=self.merge_global_params(table_attributes=table_attributes),
            ignore_params=table_attributes.IgnoreParameters,
        )
        query = Query(
            query_str=query_str,
            default_table_prefix=template.default_table_prefix,
        )

        downstairs: str = table_attributes.TableName
        mapping_labels: Dict[str, Any] = table_attributes.Labels
        metadata: List[Dict[str, Any]] = self._mapping_config.Metadata

        if downstairs not in self.mapped:
            self.mapped[downstairs] = {}

        for upstairs_attributes in query.detect_upstairs_attributes():
            upstairs: str = upstairs_attributes[MapKey.TABLE_NAME]

            if not self.mapped[downstairs].get(upstairs):
                self.mapped[downstairs][upstairs] = self.create_upstairs_value(
                    template=template,
                    mapping_labels=mapping_labels,
                    metadata=metadata,
                    upstairs=upstairs,
                )

            self.mapped[downstairs][upstairs][MapKey.LINES].append(
                {
                    MapKey.LINE_NUMBER: upstairs_attributes[MapKey.LINE_NUMBER],
                    MapKey.LINE_STRING: upstairs_attributes[MapKey.LINE_STRING],
                }
            )

    def get_global_params(self) -> Dict[str, Any]:
        """get global parameters in mapping.yaml

        Returns:
            dict: global parameters
        """
        _global: MappingConfigGlobal = self._mapping_config.get_global()
        return _global.Parameters

    def merge_global_params(
        self, table_attributes: MappingConfigMappingTable
    ) -> Dict[str, Any]:
        """return a combination of global parameters and table parameters

        Args:
            table_attributes (MappingConfigMappingTable): table attributes

        Returns:
            dict: combined parameters
        """
        global_params: Dict[str, Any] = self.get_global_params()
        table_params: OrderedDict[str, Any] = table_attributes.Parameters

        # Table parameters are prioritized over global parameters
        return {**global_params, **table_params}

    @staticmethod
    def create_upstairs_value(
        template: Template,
        mapping_labels: Dict[str, Any],
        metadata: List[Dict[str, Any]],
        upstairs: str,
    ) -> Dict[str, Any]:
        """create upstairs table information

        Args:
            template (Template): Template class
            mapping_labels (dict): labels in mapping section
            metadata (list[str]): metadata
            upstairs (str): upstairs table name

        Returns:
            dict: upstairs table information
        """
        metadata_labels: List[Dict[str, Any]] = []
        upstairs_values = {
            MapKey.TEMPLATE_SOURCE_TYPE: template.source_type.value,
            MapKey.KEY: template.key,
            MapKey.URI: template.uri,
            MapKey.LINES: [],
        }

        if template.source_type in (TemplateSourceType.GCS, TemplateSourceType.S3):
            upstairs_values[MapKey.BUCKET_NAME] = template.bucket
        elif template.source_type == TemplateSourceType.REDASH:
            upstairs_values[MapKey.DATA_SOURCE_NAME] = template.data_source_name

        if metadata:
            metadata_labels = [
                m.get(MappingConfigKey.LABELS, {})
                for m in metadata
                if m.get(MappingConfigKey.TABLE_NAME) == upstairs
            ]
        if mapping_labels or metadata_labels:
            upstairs_values[MapKey.LABELS] = {}

        if mapping_labels:
            upstairs_values[MapKey.LABELS] = {
                **upstairs_values[MapKey.LABELS],
                **mapping_labels,
            }

        if metadata_labels:
            upstairs_values[MapKey.LABELS] = {
                **upstairs_values[MapKey.LABELS],
                **metadata_labels[0],
            }
        return upstairs_values

    def add_unmapped_params(self, template: Template, params: List[str] = None) -> None:
        """add to the list of unmapped params

        Args:
            template (Template): SQL template
            params (dict, optional): Jinja parameters
        """
        if not params:
            template_str = template.get_template_str()
            params = template.detect_jinja_params(template_str=template_str)
        self.unmapped.append(
            {
                MapKey.TEMPLATE: template,
                MapKey.PARAMETERS: params,
            }
        )

    def detect_unmapped_params(
        self, template: Template, table_attributes: MappingConfigMappingTable
    ) -> List[str]:
        """detect unmapped parameters in mapped files

        Args:
            template (Template): SQL template
            table_attributes (MappingConfigMappingTable):
                Table attributes from mapping configuration
        """
        template_str: str = template.get_template_str()
        template_params: List[str] = template.detect_jinja_params(template_str)
        if not template_params:
            return []

        mapped_params_dict: Dict[str, Any] = self.merge_global_params(
            table_attributes=table_attributes
        )
        mapped_params: List[str] = create_dict_key_list(d=mapped_params_dict)
        ignore_params: List[str] = table_attributes.IgnoreParameters
        unmapped_params: List[str] = list(
            set(template_params) - set(mapped_params) - set(ignore_params)
        )

        return unmapped_params


def create_dict_key_list(d: Dict[str, Any], delimiter: str = ".") -> List[str]:
    """combine nested dictionary keys and converts to a list

    Args:
        d (dict): dict[str, Any]

    Returns:
        list: results
    """
    results = []
    for key, value in d.items():
        if isinstance(value, dict):
            recursive_results = create_dict_key_list(d=value)
            for recursive_result in recursive_results:
                concat = key + delimiter + recursive_result
                results.append(concat)
        else:
            results.append(key)
    return results



import enum
import json
from logging import getLogger
from typing import Any, Dict, List, Union

from .configurator import (
    MAPPING_CONFIG_PREFIX_DEFAULT,
    STAIRLIGHT_CONFIG_PREFIX_DEFAULT,
    Configurator,
)
from .map import Map
from .source.config import (
    MapKey,
    MappingConfig,
    StairlightConfig,
    StairlightConfigSettings,
)
from .source.controller import LoadMapController, SaveMapController

logger = getLogger(__name__)


class ResponseType(enum.Enum):
    """Enum: Execution result type of up/down command"""

    TABLE: str = "table"
    FILE: str = "file"

    def __str__(self):
        return self.name


class SearchDirection(enum.Enum):
    """Enum: Search direction"""

    UP: str = "Upstairs"
    DOWN: str = "Downstairs"

    def __str__(self):
        return self.name


class Node:
    """A Node of singly-linked list"""

    def __init__(self, val: str, next: "Node" = None):
        self.val: str = val
        self.next: "Node" = None


class StairLight:
    """Table dependency detector"""

    def __init__(
        self,
        config_dir: str = ".",
        load_files: List[str] = None,
        save_file: str = "",
    ) -> None:
        """Table dependency detector

        Args:
            config_dir (str, optional):
                Stairlight configuration directory. Defaults to ".".
            load_files (list, optional):
                file names of loading results if load option set. Defaults to None.
            save_file (str, optional):
                A file name of saving results if save option set. Defaults to None.
        """
        self.load_files = load_files
        self.save_file: str = save_file
        self._configurator = Configurator(dir=config_dir)
        self._mapped: Dict[str, Any] = {}
        self._unmapped: List[Dict[str, Any]] = []
        self._mapping_config: MappingConfig = None
        self._stairlight_config: StairlightConfig = self._configurator.read_stairlight(
            prefix=STAIRLIGHT_CONFIG_PREFIX_DEFAULT
        )

    def create_map(self) -> None:
        if not self._stairlight_config:
            return

        if self.load_files:
            self.load_map()
        else:
            self._set_config()
            self._write_map()

        if self.save_file:
            self.save_map()

    @property
    def mapped(self) -> Dict[str, Any]:
        """Return mapped

        Returns:
            dict: Mapped results
        """
        return self._mapped

    @property
    def unmapped(self) -> List[Dict[str, Any]]:
        """Return unmapped

        Returns:
            list[dict]: Unmapped results
        """
        return self._unmapped

    def has_stairlight_config(self) -> bool:
        """Exists stairlight configuration file or not

        Returns:
            bool: Exists stairlight configuration file or not
        """
        return self._stairlight_config is not None

    def _set_config(self) -> None:
        """Set config"""
        if not self._stairlight_config:
            logger.warning(f"{STAIRLIGHT_CONFIG_PREFIX_DEFAULT}.y(a)ml' is not found.")
            return

        mapping_config_prefix: str = MAPPING_CONFIG_PREFIX_DEFAULT
        if self._stairlight_config.Settings:
            settings: StairlightConfigSettings = StairlightConfigSettings(
                **self._stairlight_config.Settings
            )
            if settings.MappingPrefix:
                mapping_config_prefix = settings.MappingPrefix
        self._mapping_config = self._configurator.read_mapping(
            prefix=mapping_config_prefix
        )

    def _write_map(self) -> None:
        """write a dependency map"""
        dependency_map = Map(
            stairlight_config=self._stairlight_config,
            mapping_config=self._mapping_config,
        )

        dependency_map.write()
        if self._mapping_config:
            self._mapped = dependency_map.mapped

        self._unmapped = dependency_map.unmapped

    def init(self, prefix: str = STAIRLIGHT_CONFIG_PREFIX_DEFAULT) -> str:
        """Create Stairlight template file

        Args:
            prefix (str, optional):
                Template file prefix. Defaults to STAIRLIGHT_CONFIG_PREFIX.

        Returns:
            str: Template file name
        """
        return self._configurator.create_stairlight_file(prefix=prefix)

    def check(self, prefix: str = MAPPING_CONFIG_PREFIX_DEFAULT) -> str:
        """Check mapped results and create a mapping template file

        Args:
            prefix (str, optional):
                Template file prefix. Defaults to MAPPING_CONFIG_PREFIX.

        Returns:
            str: Template file name
        """
        if self.load_files:
            logger.warning("Load option is used, skip checking.")
            return ""
        elif not self._unmapped:
            return ""

        return self._configurator.create_mapping_file(
            unmapped=self._unmapped, prefix=prefix
        )

    def save_map(self) -> None:
        """Save mapped results"""
        save_map_controller = SaveMapController(
            save_file=self.save_file, mapped=self._mapped
        )
        save_map_controller.save()

    def load_map(self) -> None:
        """Load mapped results"""
        if not self.load_files:
            return
        for load_file in self.load_files:
            load_map_controller = LoadMapController(load_file=load_file)
            loaded_map = load_map_controller.load()

            if self._mapped:
                self._mapped = deep_merge(org=self._mapped, add=loaded_map)
            else:
                self._mapped = loaded_map

    def up(
        self,
        table_name: str,
        recursive: bool = False,
        verbose: bool = False,
        response_type: str = ResponseType.TABLE.value,
    ) -> Union[List[str], Dict[str, Any]]:
        """Search upstream nodes

        Args:
            table_name (str): Table name
            recursive (bool, optional): Search recursively or not. Defaults to False.
            verbose (bool, optional): Return verbose results or not. Defaults to False.
            response_type (str, optional):
                Response type. Defaults to ResponseType.TABLE.value.

        Returns:
            Union[list[str], dict]: Search results
        """
        return self.search(
            table_name=table_name,
            recursive=recursive,
            verbose=verbose,
            response_type=response_type,
            direction=SearchDirection.UP,
        )

    def down(
        self,
        table_name,
        recursive=False,
        verbose=False,
        response_type=ResponseType.TABLE.value,
    ) -> Union[List[str], Dict[str, Any]]:
        """Search downstream nodes

        Args:
            table_name (str): Table name
            recursive (bool, optional): Search recursively or not. Defaults to False.
            verbose (bool, optional): Return verbose results or not. Defaults to False.
            response_type (str, optional):
                Response type. Defaults to ResponseType.TABLE.value.

        Returns:
            Union[list[str], dict]: Search results
        """
        return self.search(
            table_name=table_name,
            recursive=recursive,
            verbose=verbose,
            response_type=response_type,
            direction=SearchDirection.DOWN,
        )

    def search(
        self,
        table_name: str,
        recursive: bool,
        verbose: bool,
        response_type: str,
        direction: SearchDirection,
    ) -> Union[List[str], Dict[str, Any]]:
        """Search nodes

        Args:
            table_name (str): Table name
            recursive (bool): Search recursively or not
            verbose (bool): Return verbose results or not
            response_type (str): Response type value
            direction (SearchDirection): Search direction

        Returns:
            Union[list[str], dict]: Search results
        """
        if verbose:
            return self.search_verbose(
                table_name=table_name,
                recursive=recursive,
                direction=direction,
                searched_tables=[],
                head=True,
            )

        if response_type in [type.value for type in ResponseType]:
            return self.search_plain(
                table_name=table_name,
                recursive=recursive,
                response_type=response_type,
                direction=direction,
                searched_tables=[],
                head=True,
            )

        return []

    def search_verbose(
        self,
        table_name: str,
        recursive: bool,
        direction: SearchDirection,
        searched_tables: List[str],
        head: bool,
    ) -> dict:
        """Search nodes and return verbose results

        Args:
            table_name (str): Table name
            recursive (bool): Search recursively or not
            direction (SearchDirection): Search direction
            searched_tables (list[str]): a list of searched tables
            head (bool): Current position is head or not

        Returns:
            dict: Search results
        """
        relative_map = self.create_relative_map(
            table_name=table_name, direction=direction
        )
        response: Dict[str, Any] = {table_name: {}}
        if not relative_map:
            return response

        if recursive:
            for next_table_name in relative_map.keys():
                if head:
                    searched_tables = [table_name]

                searched_tables.append(next_table_name)

                if is_cyclic(tables=searched_tables):
                    details = {
                        "table_name": table_name,
                        "next_table_name": next_table_name,
                        "searched_tables": searched_tables,
                    }
                    logger.warning(f"Circular reference detected!: {details}")
                    continue

                next_response = self.search_verbose(
                    table_name=next_table_name,
                    direction=direction,
                    recursive=recursive,
                    searched_tables=searched_tables,
                    head=False,
                )

                if not next_response.get(next_table_name):
                    continue

                relative_map[next_table_name] = {
                    **relative_map[next_table_name],
                    **next_response[next_table_name],
                }

        response[table_name][direction.value] = relative_map
        return response

    def search_plain(
        self,
        table_name: str,
        recursive: bool,
        response_type: str,
        direction: SearchDirection,
        searched_tables: List[str],
        head: bool,
    ) -> List[str]:
        """Search nodes and return simple results

        Args:
            table_name (str): Table name
            recursive (bool): Search recursively or not
            response_type (str): Response type value
            direction (SearchDirection): Search direction
            searched_tables (list[str]): a list of searched tables
            head (bool): Current position is head or not

        Returns:
            list[str]: Search results
        """
        relative_map = self.create_relative_map(
            table_name=table_name, direction=direction
        )
        response: List[str] = []
        if not relative_map:
            return response

        next_table_name: str
        for next_table_name in relative_map.keys():
            if recursive:
                if head:
                    searched_tables = []
                    searched_tables.append(table_name)

                searched_tables.append(next_table_name)

                if is_cyclic(tables=searched_tables):
                    details = {
                        "table_name": table_name,
                        "next_table_name": next_table_name,
                        "searched_tables": searched_tables,
                    }
                    logger.info(f"Circular reference detected!: {details}")
                    continue

                next_response = self.search_plain(
                    table_name=next_table_name,
                    direction=direction,
                    recursive=recursive,
                    response_type=response_type,
                    searched_tables=searched_tables,
                    head=False,
                )
                response = response + next_response

            if response_type == ResponseType.TABLE.value:
                response.append(next_table_name)
            elif response_type == ResponseType.FILE.value:
                response.append(relative_map[next_table_name].get(MapKey.URI))
            logger.debug(json.dumps(response, indent=2))

        return sorted(list(set(response)))

    def create_relative_map(
        self, table_name: str, direction: SearchDirection
    ) -> Dict[str, Any]:
        """Create a relative map for the specified direction

        Args:
            table_name (str): Table name
            direction (SearchDirection): Search direction

        Returns:
            dict: Relative map
        """
        relative_map: Dict[str, Any] = {}
        if direction == SearchDirection.UP:
            relative_map = self._mapped.get(table_name, {})
        elif direction == SearchDirection.DOWN:
            for key in [k for k, v in self._mapped.items() if v.get(table_name)]:
                relative_map[key] = self._mapped[key][table_name]
        return relative_map

    def find_tables_by_labels(self, target_labels: List[str]) -> List[str]:
        """Find tables by labels

        Args:
            target_labels (list[str]): Target labels

        Returns:
            list[str]: Tables to search
        """
        tables_to_search: List[str] = []

        # "mapping" section in mapping.yaml
        for mapping in self._mapping_config.get_mapping():
            for mapping_table in mapping.get_table():
                if (
                    mapping_table.TableName not in tables_to_search
                    and self.is_target_label_found(
                        target_labels=target_labels,
                        configured_labels=mapping_table.Labels,
                    )
                ):
                    tables_to_search.append(mapping_table.TableName)

        # "metadata" section in mapping.yaml
        for metadata in self._mapping_config.get_metadata():
            if (
                metadata.TableName not in tables_to_search
                and self.is_target_label_found(
                    target_labels=target_labels,
                    configured_labels=metadata.Labels,
                )
            ):
                tables_to_search.append(metadata.TableName)

        return tables_to_search

    @staticmethod
    def is_target_label_found(
        target_labels: List[str], configured_labels: Dict[str, Any]
    ) -> bool:
        found_count: int = 0
        configured_label_key: str
        configured_label_value: str
        for configured_label_key, configured_label_value in configured_labels.items():
            for target_label in target_labels:
                target_label_key = target_label.split(":")[0]
                target_label_value = target_label.split(":")[1]
                if (
                    target_label_key == configured_label_key
                    and target_label_value == configured_label_value
                ):
                    found_count += 1

        return found_count == len(target_labels)


def is_cyclic(tables: List[str]) -> bool:
    """Floyd's cycle-finding algorithm

    Args:
        tables (list[str]): Detected tables

    Returns:
        bool: Table dependencies are cyclic or not
    """
    nodes: Dict[str, Node] = {}
    for table in tables:
        if table not in nodes.keys():
            nodes[table] = Node(table)
    for i, table in enumerate(tables):
        if not nodes[table].next and i < len(tables) - 1:
            nodes[table].next = nodes[tables[i + 1]]
    slow = fast = nodes[tables[0]]
    while fast and fast.next:
        slow = slow.next
        fast = fast.next.next
        if slow == fast:
            return True
    return False


def deep_merge(org: Dict[str, Any], add: Dict[str, Any]) -> Dict[str, Any]:
    """Merge nested dicts

    Args:
        org (dict[str, Any]): original dict
        add (dict[str, Any]): dict to add

    Returns:
        dict: merged dict
    """
    new: Dict[str, Any] = org
    for add_key, add_value in add.items():
        org_value: Dict[str, Any] = org.get(add_key, {})
        if add_key not in org:
            new[add_key] = add_value
        elif isinstance(add_value, dict):
            new[add_key] = deep_merge(org=org_value, add=add_value)
    return new



import os
from dataclasses import asdict
from logging import getLogger
from typing import Any, Dict, Iterator, List

from sqlalchemy import create_engine, text
from sqlalchemy.engine.row import Row

from ..config import MappingConfig, MappingConfigMappingTable, StairlightConfig
from ..config_key import StairlightConfigKey
from ..template import Template, TemplateSource, TemplateSourceType
from .config import StairlightConfigIncludeRedash

logger = getLogger(__name__)


class RedashTemplate(Template):
    def __init__(
        self,
        mapping_config: MappingConfig,
        query_id: int,
        query_name: str,
        query_str: str = "",
        data_source_name: str = "",
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=str(query_id),
            source_type=TemplateSourceType.REDASH,
        )
        self.query_id = query_id
        self.query_str = query_str
        self.uri = query_name
        self.data_source_name = data_source_name

    def find_mapped_table_attributes(self) -> Iterator[MappingConfigMappingTable]:
        """Get mapped tables as iterator

        Yields:
            Iterator[dict]: Mapped table attributes
        """
        mapping: Any
        for mapping in self._mapping_config.get_mapping():
            if mapping.TemplateSourceType != self.source_type.value:
                continue

            if (
                mapping.DataSourceName == self.data_source_name
                and mapping.QueryId == self.query_id
            ):
                for table_attributes in mapping.get_table():
                    yield table_attributes
                break

    def get_template_str(self) -> str:
        """Get template string that read from Redash
        Returns:
            str: Template string
        """
        return self.query_str


class RedashTemplateSource(TemplateSource):
    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        include: StairlightConfigIncludeRedash,
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config,
            mapping_config=mapping_config,
        )
        self._include = include
        self.where_clause: List[str] = []
        self.conditions: Dict[str, Any] = self.make_conditions()

    def make_conditions(self) -> Dict[str, Any]:
        query_ids = self._include.QueryIds
        return {
            StairlightConfigKey.Redash.DATA_SOURCE_NAME: {
                "key": StairlightConfigKey.Redash.DATA_SOURCE_NAME,
                "query": "data_sources.name = :data_source",
                "parameters": self._include.DataSourceName,
            },
            StairlightConfigKey.Redash.QUERY_IDS: {
                "key": StairlightConfigKey.Redash.QUERY_IDS,
                "query": "queries.id IN :query_ids",
                "parameters": (tuple(query_ids) if query_ids else None),
            },
        }

    def search_templates(self) -> Iterator[Template]:
        results = self.get_queries_from_redash()
        for result in results:
            yield RedashTemplate(
                mapping_config=self._mapping_config,
                query_id=result[0],
                query_name=result[1],
                query_str=result[2],
                data_source_name=result[3],
            )

    def get_queries_from_redash(self) -> List[Row]:
        sql_file_name = "sql/redash_queries.sql"
        current_dir = os.path.dirname(os.path.abspath(__file__))
        query_text = text(
            self.build_query_string(path=f"{current_dir}/{sql_file_name}")
        )

        data_source_condition: Dict[str, Any] = self.conditions.get(
            StairlightConfigKey.Redash.DATA_SOURCE_NAME, {}
        )
        query_ids_condition: Dict[str, Any] = self.conditions.get(
            StairlightConfigKey.Redash.QUERY_IDS, {}
        )
        connection_str = self.get_connection_str()
        engine = create_engine(connection_str)
        queries = engine.execute(
            query_text,
            data_source=data_source_condition.get("parameters", ""),
            query_ids=query_ids_condition.get("parameters", []),
        ).fetchall()

        return queries

    def build_query_string(self, path: str) -> str:
        base_query_string = self.read_query_from_file(path=path)
        for condition in self.conditions.values():
            if condition["key"] in asdict(self._include).keys():
                self.where_clause.append(condition["query"])
        return base_query_string + "WHERE " + " AND ".join(self.where_clause)

    def read_query_from_file(self, path: str) -> str:
        with open(path, "r", encoding="utf-8") as f:
            return f.read()

    def get_connection_str(self) -> str:
        environment_variable_name = self._include.DatabaseUrlEnvironmentVariable
        connection_str = os.environ.get(environment_variable_name, "")
        if not connection_str:
            logger.error(f"{environment_variable_name} is not found.")
        return connection_str



import pathlib
import re
from typing import Iterator, Optional

from ..config import ConfigAttributeNotFoundException, MappingConfig, StairlightConfig
from ..template import Template, TemplateSource, TemplateSourceType
from .config import StairlightConfigIncludeFile


class FileTemplate(Template):
    def __init__(
        self,
        mapping_config: MappingConfig,
        key: str,
        default_table_prefix: Optional[str] = None,
    ):
        super().__init__(
            mapping_config=mapping_config,
            key=key,
            source_type=TemplateSourceType.FILE,
            default_table_prefix=default_table_prefix,
        )
        self.uri = self.get_uri()

    def get_uri(self) -> str:
        """Get uri from file path

        Returns:
            str: uri
        """
        return str(pathlib.Path(self.key).resolve())

    def get_template_str(self) -> str:
        """Get template string that read from a file in local file system

        Returns:
            str: Template string
        """
        with open(self.key) as f:
            return f.read()


class FileTemplateSource(TemplateSource):
    def __init__(
        self,
        stairlight_config: StairlightConfig,
        mapping_config: MappingConfig,
        include: StairlightConfigIncludeFile,
    ) -> None:
        super().__init__(
            stairlight_config=stairlight_config,
            mapping_config=mapping_config,
        )
        self._include = include

    def search_templates(self) -> Iterator[Template]:
        """Search SQL template files from local file system

        Yields:
            Iterator[Template]: SQL template file attributes
        """
        if not self._include.FileSystemPath:
            raise ConfigAttributeNotFoundException(
                f"FileSystemPath is not found. {self._include}"
            )

        path_obj = pathlib.Path(self._include.FileSystemPath)
        for p in path_obj.glob("**/*"):
            if (
                (p.is_dir())
                or (
                    not re.fullmatch(
                        rf"{self._include.Regex}",
                        str(p),
                    )
                )
                or self.is_excluded(
                    source_type=TemplateSourceType(self._include.TemplateSourceType),
                    key=str(p),
                )
            ):
                self.logger.debug(f"{str(p)} is skipped.")
                continue

            yield FileTemplate(
                mapping_config=self._mapping_config,
                key=str(p),
                default_table_prefix=self._include.DefaultTablePrefix,
            )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['stairlight',
 'stairlight.source',
 'stairlight.source.dbt',
 'stairlight.source.file',
 'stairlight.source.gcs',
 'stairlight.source.redash',
 'stairlight.source.s3']

package_data = \
{'': ['*'], 'stairlight.source.redash': ['sql/*']}

install_requires = \
['Jinja2>=2.10.3', 'PyYAML>=5.0']

extras_require = \
{':extra == "s3"': ['boto3-stubs[s3]>=1.24.17,<2.0.0'],
 'dbt-bigquery': ['dbt-core>=1.1,<2.0', 'dbt-bigquery>=1.1,<2.0'],
 'gcs': ['google-cloud-storage>=1.28.1,<2.0.0'],
 'redash': ['psycopg2>=2.9.3,<3.0.0', 'SQLAlchemy>=1.4.31,<2.0.0'],
 's3': ['boto3>=1.24.14,<2.0.0']}

entry_points = \
{'console_scripts': ['stairlight = stairlight.cli:main']}

setup_kwargs = {
    'name': 'stairlight',
    'version': '0.6.0',
    'description': 'An end-to-end data lineage tool',
    'long_description': '<div align="center">\n  <img src="https://raw.githubusercontent.com/tosh2230/stairlight/main/img/stairlight_white.png" width="400" alt="Stairlight">\n</div>\n\n-----------------\n\n# Stairlight\n\n[![PyPi Version](https://img.shields.io/pypi/v/stairlight.svg?style=flat-square&logo=PyPi)](https://pypi.org/project/stairlight/)\n[![PyPi License](https://img.shields.io/pypi/l/stairlight.svg?style=flat-square)](https://pypi.org/project/stairlight/)\n[![PyPi Python Versions](https://img.shields.io/pypi/pyversions/stairlight.svg?style=flat-square)](https://pypi.org/project/stairlight/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square)](https://github.com/psf/black)\n[![CI](https://github.com/tosh2230/stairlight/actions/workflows/ci.yml/badge.svg)](https://github.com/tosh2230/stairlight/actions/workflows/ci.yml)\n\nAn end-to-end data lineage tool, detects table dependencies by SQL SELECT statements.\n\n<div align="left">\n  <img src="https://raw.githubusercontent.com/tosh2230/stairlight/main/img/drawio/concepts.drawio.png" width="800" alt="concepts">\n</div>\n\n## Supported Data Sources\n\n| Data Source | Remarks |\n| --- | --- |\n| Local file system | With Python Pathlib module |\n| [Amazon S3](https://aws.amazon.com/s3/) | |\n| [Google Cloud Storage](https://cloud.google.com/storage) | Also available for [Google Cloud Composer](https://cloud.google.com/composer) |\n| [dbt](https://www.getdbt.com/) - [Google BigQuery](https://cloud.google.com/bigquery) | Using `dbt compile` command internally |\n| [Redash](https://redash.io/) | |\n\n## Installation\n\nThis package is distributed on [PyPI](https://pypi.org/project/stairlight/).\n\n```sh\n# The base package is for local file system only.\n$ pip install stairlight\n\n# Set extras when detecting from other data sources.\n# e.g. Amazon S3 and Google Cloud Storage\n$ pip install "stairlight[s3, gcs]"\n```\n\n| Data Source | TemplateSourceType | Extra |\n| --- | --- | --- |\n| Local file system | File | - |\n| Amazon S3 | S3 | s3 |\n| Google Cloud Storage | GCS | gcs |\n| dbt - Google Bigquery | dbt | dbt-bigquery |\n| Redash | Redash | redash |\n\n## Getting Started\n\nThere are 3 steps to use.\n\n```sh\n# Step 1: Initialize and set data location settings\n$ stairlight init\n\'./stairlight.yaml\' has created.\nPlease edit it to set your data sources.\n\n# Step 2: Map your SQL statements and tables, and add metadata\n$ stairlight map\n\'./mapping_yyyyMMddhhmmss.yaml\' has created.\nPlease map undefined tables and parameters, and append to your latest configuration file.\n\n# Step 3: Get a table dependency map\n$ stairlight\n```\n\n## Description\n\n### Input\n\n- SQL SELECT statements\n- Configuration YAML files\n    - stairlight.yaml: SQL statements locations and include/exclude conditions.\n    - mapping.yaml: Mapping SQL statements and tables.\n\n### Output\n\n- JSON dependency map\n\n    <details>\n\n    <summary>Example</summary>\n\n    ```json\n    {\n      "PROJECT_d.DATASET_e.TABLE_f": {\n        "PROJECT_j.DATASET_k.TABLE_l": {\n          "TemplateSourceType": "File",\n          "Key": "tests/sql/main/one_line_2.sql",\n          "Uri": "/foo/bar/stairlight/tests/sql/main/one_line_2.sql",\n          "Lines": [\n            {\n              "LineNumber": 1,\n              "LineString": "SELECT * FROM PROJECT_j.DATASET_k.TABLE_l WHERE 1 = 1"\n            }\n          ]\n        },\n        "PROJECT_C.DATASET_C.TABLE_C": {\n          "TemplateSourceType": "GCS",\n          "Key": "sql/cte/cte_multi_line.sql",\n          "Uri": "gs://stairlight/sql/cte/cte_multi_line.sql",\n          "Lines": [\n            {\n              "LineNumber": 6,\n              "LineString": "        PROJECT_C.DATASET_C.TABLE_C"\n            }\n          ],\n          "BucketName": "stairlight",\n          "Labels": {\n            "Source": "gcs",\n            "Test": "b"\n          }\n        }\n      },\n      "AggregateSales": {\n        "PROJECT_e.DATASET_e.TABLE_e": {\n          "TemplateSourceType": "Redash",\n          "Key": 5,\n          "Uri": "AggregateSales",\n          "Lines": [\n            {\n              "LineNumber": 1,\n              "LineString": "SELECT service, SUM(total_amount) FROM PROJECT_e.DATASET_e.TABLE_e GROUP BY service"\n            }\n          ],\n          "DataSourceName": "BigQuery",\n          "Labels": {\n            "Category": "Sales"\n          }\n        }\n      },\n      "dummy.dummy.example_b": {\n        "PROJECT_t.DATASET_t.TABLE_t": {\n          "TemplateSourceType": "dbt",\n          "Key": "tests/dbt/project_01/target/compiled/project_01/models/b/example_b.sql",\n          "Uri": "/foo/bar/stairlight/tests/dbt/project_01/target/compiled/project_01/models/b/example_b.sql",\n          "Lines": [\n            {\n              "LineNumber": 1,\n              "LineString": "select * from PROJECT_t.DATASET_t.TABLE_t where value_a = 0 and value_b = 0"\n            }\n          ]\n        }\n      },\n      "PROJECT_as.DATASET_bs.TABLE_cs": {\n        "PROJECT_A.DATASET_A.TABLE_A": {\n          "TemplateSourceType": "S3",\n          "Key": "sql/one_line/one_line.sql",\n          "Uri": "s3://stairlight/sql/one_line/one_line.sql",\n          "Lines": [\n            {\n              "LineNumber": 1,\n              "LineString": "SELECT * FROM PROJECT_A.DATASET_A.TABLE_A WHERE 1 = 1"\n            }\n          ],\n          "BucketName": "stairlight",\n          "Labels": {\n            "Source": null,\n            "Test": "a"\n          }\n        }\n      }\n    }\n    ```\n\n    </details>\n\n### Collecting patterns\n\n#### Centralization\n\n<div align="left">\n  <img src="https://raw.githubusercontent.com/tosh2230/stairlight/main/img/drawio/centralization.drawio.png" width="800" alt="centralization">\n</div>\n\n#### Agents\n\n<div align="left">\n  <img src="https://raw.githubusercontent.com/tosh2230/stairlight/main/img/drawio/agents.drawio.png" width="800" alt="agents">\n</div>\n\n## Configuration\n\nConfiguration files can be found [here](https://github.com/tosh2230/stairlight/tree/main/tests/config), used for unit testing in CI.\n\n### stairlight.yaml\n\n\'stairlight.yaml\' is for setting up Stairlight itself. It is responsible for specifying SQL statements to be read.\n\n`init` command creates a template of stairlight.yaml.\n\n  <details>\n\n  <summary>Example</summary>\n\n  ```yaml\n  Include:\n    - TemplateSourceType: File\n      FileSystemPath: "./tests/sql"\n      Regex: ".*/*.sql$"\n      DefaultTablePrefix: "PROJECT_A"\n    - TemplateSourceType: GCS\n      ProjectId: null\n      BucketName: stairlight\n      Regex: "^sql/.*/*.sql$"\n      DefaultTablePrefix: "PROJECT_A"\n    - TemplateSourceType: Redash\n      DatabaseUrlEnvironmentVariable: REDASH_DATABASE_URL\n      DataSourceName: BigQuery\n      QueryIds:\n        - 1\n        - 3\n        - 5\n    - TemplateSourceType: dbt\n      ProjectDir: tests/dbt/project_01\n      ProfilesDir: tests/dbt\n      Vars:\n        key_a: value_a\n        key_b: value_b\n    - TemplateSourceType: S3\n      BucketName: stairlight\n      Regex: "^sql/.*/*.sql$"\n      DefaultTablePrefix: "PROJECT_A"\n  Exclude:\n    - TemplateSourceType: File\n      Regex: "main/exclude.sql$"\n  Settings:\n    MappingPrefix: "mapping"\n  ```\n\n  </details>\n\n### mapping.yaml\n\n\'mapping.yaml\' is used to define relationships between input SELECT statements and table names.\n\n`map` command creates a template of mapping.yaml, based on the configuration of stairlight.yaml.\n\n  <details>\n\n  <summary>Example</summary>\n\n  ```yaml\n  Global:\n    Parameters:\n      DESTINATION_PROJECT: stairlight\n      params:\n        PROJECT: 1234567890\n        DATASET: public\n        TABLE: taxirides\n  Mapping:\n    - TemplateSourceType: File\n      FileSuffix: "tests/sql/main/union_same_table.sql"\n      Tables:\n        - TableName: "test_project.beam_streaming.taxirides_aggregation"\n          Parameters:\n            params:\n              source_table: source\n              destination_table: destination\n          IgnoreParameters:\n            - execution_date.add(days=1).isoformat()\n    - TemplateSourceType: GCS\n      Uri: "gs://stairlight/sql/one_line/one_line.sql"\n      Tables:\n        - TableName: "PROJECT_a.DATASET_b.TABLE_c"\n    - TemplateSourceType: Redash\n      QueryId: 5\n      DataSourceName: metadata\n      Tables:\n        - TableName: New Query\n          Parameters:\n            table: dashboards\n          Labels:\n            Category: Redash test\n    - TemplateSourceType: dbt\n      ProjectName: project_01\n      FileSuffix: tests/dbt/project_01/target/compiled/project_01/models/example/my_first_dbt_model.sql\n      Tables:\n        - TableName: dummy.dummy.my_first_dbt_model\n    - TemplateSourceType: S3\n      Uri: "s3://stairlight/sql/one_line/one_line.sql"\n      Tables:\n        - TableName: "PROJECT_as.DATASET_bs.TABLE_cs"\n  Metadata:\n    - TableName: "PROJECT_A.DATASET_A.TABLE_A"\n      Labels:\n        Source: Null\n        Test: a\n  ```\n\n  </details>\n\n#### Global Section\n\nThis section is for global configurations.\n\n`Parameters` attribute is used to set common parameters. If conflicts has occurred with `Parameters` attributes in mapping section, mapping section\'s parameters will be used in preference to global.\n\n#### Mapping Section\n\nMapping section is used to define relationships between input SELECT statements and tables that created as a result of query execution.\n\n`Parameters` attribute allows you to reflect settings in [jinja](https://jinja.palletsprojects.com/) template variables embedded in statements. If multiple settings are applied to a statement using jinja template, the statement will be read as if there were the same number of queries as the number of settings.\n\nIn contrast, `IgnoreParameters` attribute handles a list to ignore when rendering queries.\n\n#### Metadata Section\n\nThis section is mainly used to set metadata to tables appears only in queries.\n\n## Arguments and Options\n\n```txt\n$ stairlight --help\nusage: stairlight [-h] [-c CONFIG] [--save SAVE] [--load LOAD] {init,check,up,down} ...\n\nAn end-to-end data lineage tool, detects table dependencies by SQL SELECT statements.\nWithout positional arguments, return a table dependency map as JSON format.\n\npositional arguments:\n  {init,map,check,up,down}\n    init                create new Stairlight configuration file\n    map (check)         create new configuration file about undefined mappings\n    up                  return upstairs ( table | SQL file ) list\n    down                return downstairs ( table | SQL file ) list\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set Stairlight configuration directory\n  -q, --quiet           keep silence\n  --save SAVE           A file path where map results will be saved.\n                        You can choose from local file system, GCS, S3.\n  --load LOAD           A file path where map results are saved.\n                        You can choose from local file system, GCS, S3.\n                        It can be specified multiple times.\n```\n\n### init\n\n`init` creates a new Stairlight configuration file.\n\n```txt\n$ stairlight init --help\nusage: stairlight init [-h] [-c CONFIG]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set Stairlight configuration directory.\n  -q, --quiet           keep silence\n```\n\n### map(check)\n\n`map` creates new configuration file about undefined mappings.`check` is an alias.\nThe option specification is the same as `init`.\n\n### up\n\n`up` outputs a list of tables or SQL files located upstream from the specified table.\n\n- Use table(`-t`, `--table`) or label(`-l`, `--label`) option to specify tables to search.\n- Recursive option(`-r`, `--recursive`) is set, Stairlight will find tables recursively and output as a list.\n- Verbose option(`-v`, `--verbose`) is set, Stairlight will add detailed information and output it as a dict.\n\n```txt\n$ stairlight up --help\nusage: stairlight up [-h] [-c CONFIG] [--save SAVE] [--load LOAD] (-t TABLE | -l LABEL) [-o {table,file}]\n                     [-v] [-r]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -c CONFIG, --config CONFIG\n                        set Stairlight configuration directory\n  -q, --quiet           keep silence\n  --save SAVE           A file path where map results will be saved.\n                        You can choose from local file system, GCS, S3.\n  --load LOAD           A file path where map results are saved.\n                        You can choose from local file system, GCS, S3.\n                        It can be specified multiple times.\n  -t TABLE, --table TABLE\n                        table names that Stairlight searches for, can be specified\n                        multiple times. e.g. -t PROJECT_a.DATASET_b.TABLE_c -t\n                        PROJECT_d.DATASET_e.TABLE_f\n  -l LABEL, --label LABEL\n                        labels set for the table in mapping configuration, can be specified multiple times.\n                        The separator between key and value should be a colon(:).\n                        e.g. -l key_1:value_1 -l key_2:value_2\n  -o {table,file}, --output {table,file}\n                        output type\n  -v, --verbose         return verbose results\n  -r, --recursive       search recursively\n```\n\n### down\n\n`down` outputs a list of tables or SQL files located downstream from the specified table.\nThe option specification is the same as `up`.\n\n## Use as a library\n\nStairlight can also be used as a library.\n\n[tosh2230/stairlight-app](https://github.com/tosh2230/stairlight-app) is a sample web application rendering table dependency graph with Stairlight, using Graphviz, Streamlit and Google Cloud Run.\n',
    'author': 'tosh2230',
    'author_email': 'rev.to12@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)
